#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//структура, що містить дані про книги
struct book {
    char author[50];
    char title[100];
    int year;
    int pages;
    float cost;
};

//------------------------------------------------------------------------------------------

//однозв'язний список
struct ListNode {
    struct book data;
    struct ListNode* next;
};

//------------------------------------------------------------------------------------------

//конвертація масиву структур з даними в одзнозв'язний список
void appendNode(struct ListNode** head, struct book data) {
    struct ListNode* newNode = (struct ListNode*)malloc(sizeof(struct ListNode));
    newNode->data = data;
    newNode->next = NULL;

    if (*head == NULL) {
        *head = newNode;
    } else {
        struct ListNode* last = *head;
        while (last->next != NULL) {
            last = last->next;
        }
        last->next = newNode;
    }
}

//------------------------------------------------------------------------------------------

//виведення даних у вигляді таблиці
void printList(struct ListNode* head) {
    printf("| %-6s | %-23s | %-30s | %-4s | %-6s | %-8s |\n", "Index", "Author", "Book name", "Year", "Page", "Price, $");
    printf("|--------|-------------------------|--------------------------------|------|--------|----------|\n");
    int index = 1;
    while (head != NULL) {
        printf("| %-6d | %-23s | %-30s | %-4d | %-6d | %-8.2f |\n",
               index, head->data.author, head->data.title, head->data.year, head->data.pages, head->data.cost);
        head = head->next;
        index++;
    }
}

//------------------------------------------------------------------------------------------

//звільнення пам'яті від однозв'язних списків
void freeList(struct ListNode* head) {
    struct ListNode* current = head;
    struct ListNode* next;

    // Перевірка, чи не є голова списку рівною NULL
    if (head == NULL)
        return;

    while (current->next != NULL) {
        next = current->next;
        free(current);
        current = next;
    }
    free(current);
    head = NULL;
}


//------------------------------------------------------------------------------------------

//видалення книги за введеним індексом
void deleteBook(struct ListNode** head, int index) {
    if (index < 0) {
        printf("Invalid index. No book deleted.\n");
        return;
    }

    struct ListNode* current = *head;
    struct ListNode* prev = NULL;
    int i = 0;

    while (current != NULL && i < index) {
        prev = current;
        current = current->next;
        i++;
    }

    if (current == NULL) {
        printf("Invalid index. No book deleted.\n");
        return;
    }

    if (prev == NULL) {
        *head = current->next;
    } else {
        prev->next = current->next;
    }

    free(current);
    printf("Book at index %d deleted successfully!\n", index);
}

//------------------------------------------------------------------------------------------

//функція для видалення книжок в котрих менше 90 сторінок та написані до 1980 року
void deleteBooksLessThan90PagesBefore1980(struct ListNode** head) {
    struct ListNode* current = *head;
    struct ListNode* prev = NULL;

    while (current != NULL) {
        if (current->data.year < 1980 && current->data.pages < 90) {
            if (prev == NULL) {
                *head = current->next;
                free(current);
                current = *head;
            } else {
                prev->next = current->next;
                free(current);
                current = prev->next;
            }
        } else {
            prev = current;
            current = current->next;
        }
    }
}

//------------------------------------------------------------------------------------------

//збереження даних у файл
void writeBooksToFile(const char* filename, struct book books[], int num_books) {
    FILE* file = fopen(filename, "w");
    if (file == NULL) {
        printf("File opening failed.\n");
        return;
    }

    for (int i = 0; i < num_books; i++) {
        fprintf(file, "%s, %s, %d, %d, %.2f\n", books[i].author, books[i].title, books[i].year, books[i].pages, books[i].cost);
    }

    fclose(file);
}

//------------------------------------------------------------------------------------------

//зчитування даних з файлу
int readBooksFromFile(const char* filename, struct book books[], int max_books) {
    FILE* file = fopen(filename, "r");
    if (file == NULL) {
        printf("File opening failed.\n");
        return 0;
    }

    char line[256];
    int num_books = 0;

    while (fgets(line, sizeof(line), file) != NULL) {
        char author[50], title[100];
        int year, pages;
        float cost;

        sscanf(line, "%[^,], %[^,], %d, %d, %f", author, title, &year, &pages, &cost);

        strcpy(books[num_books].author, author);
        strcpy(books[num_books].title, title);
        books[num_books].year = year;
        books[num_books].pages = pages;
        books[num_books].cost = cost;

        num_books++;
        if (num_books >= max_books) {
            printf("Too many books.\n");
            break;
        }
    }

    fclose(file);
    return num_books;
}

//------------------------------------------------------------------------------------------

//сортування книжок за автором (в алфавітному порядку)
void sortByAuthor(struct ListNode** head) {
    if (*head == NULL || (*head)->next == NULL) {
        return;
    }

    struct ListNode* current = *head;
    struct ListNode* index = NULL;
    struct book temp;

    while (current != NULL) {
        index = current->next;

        while (index != NULL) {
            if (strcmp(current->data.author, index->data.author) > 0) {
                temp = current->data;
                current->data = index->data;
                index->data = temp;
            }
            index = index->next;
        }
        current = current->next;
    }
}

//------------------------------------------------------------------------------------------

//сортування книжок за назвою (в алфавітному порядку)
void sortByTitle(struct ListNode** head) {
    if (*head == NULL || (*head)->next == NULL) {
        return;
    }

    struct ListNode* current = *head;
    struct ListNode* index = NULL;
    struct book temp;

    while (current != NULL) {
        index = current->next;

        while (index != NULL) {
            if (strcmp(current->data.title, index->data.title) > 0) {
                temp = current->data;
                current->data = index->data;
                index->data = temp;
            }
            index = index->next;
        }
        current = current->next;
    }
}

//------------------------------------------------------------------------------------------

//сортування книжок за роком написання (від найстарішої до найновішої)
void sortByYear(struct ListNode** head) {
    if (*head == NULL || (*head)->next == NULL) {
        return;
    }

    struct ListNode* current = *head;
    struct ListNode* index = NULL;
    struct book temp;

    while (current != NULL) {
        index = current->next;

        while (index != NULL) {
            if (current->data.year > index->data.year) {
                temp = current->data;
                current->data = index->data;
                index->data = temp;
            }
            index = index->next;
        }
        current = current->next;
    }
}

//------------------------------------------------------------------------------------------

//сортування книжок за ціною (від найдешевших до найдорожчих)
void sortByCost(struct ListNode** head) {
    if (*head == NULL || (*head)->next == NULL) {
        return;
    }

    struct ListNode* current = *head;
    struct ListNode* index = NULL;
    struct book temp;

    while (current != NULL) {
        index = current->next;

        while (index != NULL) {
            if (current->data.cost > index->data.cost) {
                temp = current->data;
                current->data = index->data;
                index->data = temp;
            }
            index = index->next;
        }
        current = current->next;
    }
}

//------------------------------------------------------------------------------------------

//сортування за кількістю сторінок (від найкоротших до найдовших)
void sortByPages(struct ListNode** head) {
    if (*head == NULL || (*head)->next == NULL) {
        return;
    }

    struct ListNode* current = *head;
    struct ListNode* index = NULL;
    struct book temp;

    while (current != NULL) {
        index = current->next;

        while (index != NULL) {
            if (current->data.pages > index->data.pages) {
                temp = current->data;
                current->data = index->data;
                index->data = temp;
            }
            index = index->next;
        }
        current = current->next;
    }
}



int main() {
    int max_books = 100, num_books = 0;
    struct book books[max_books];
    struct ListNode* head = NULL;
    char fileName[100];
    char operationChoice;
    FILE* file;
    char opened=0;

    do {
        printf("\nChoose operation:\n\
        f - File\n\
        s - Sort\n\
        a - Add\n\
        d - Delete\n\
        q - Quit\n\
        Enter your choice: ");
        scanf(" %c", &operationChoice);

        if(operationChoice == 'f' || operationChoice == 'F'){
            printf("1 - open\n\
            2 - new\n\
            3 - save\n\
            Enter your choice: ");
            int fileChoice;
            scanf("%d", &fileChoice);
            if(fileChoice == 1){
            printf("Enter the name of the file to open: ");
            scanf("%s", fileName);
            file = fopen(fileName, "r+");
            if (file == NULL) {
                printf("File opening failed. Make sure '%s' exists.\n", fileName);
                return 1;
            } else opened=1;
            num_books = readBooksFromFile(fileName, books, max_books);
            if (num_books == 0) {
                printf("No books read from the file.\n");
                fclose(file);
                return 1;
            }
            printf("| %-6s | %-23s | %-30s | %-4s | %-6s | %-8s |\n", "Index", "Author", "Book name", "Year", "Page", "Price, $");
            printf("|--------|-------------------------|--------------------------------|------|--------|----------|\n");
            for (int i = 0; i < num_books; i++) {
                printf("| %-6d | %-23s | %-30s | %-4d | %-6d | %-8.2f |\n",
                       i+1, books[i].author, books[i].title, books[i].year, books[i].pages, books[i].cost);
            }

            printf("\nBooks by authors with names or surnames starting with 'A':\n");
            for (int i = 0; i < num_books; i++) {
                if (books[i].author[0] == 'A' || books[i].author[0] == 'a') {
                    printf("%s - %s\n", books[i].title, books[i].author);
                }
            }

            for (int i = 0; i < num_books; i++) {
                appendNode(&head, books[i]);
            }
        }
        else if(fileChoice == 2){
            char fileName[100]; // Максимальна довжина назви файлу
            printf("Enter the name of the new file: ");
            scanf("%s", fileName);
            FILE* file;
            file = fopen(fileName, "w");
            if (file == NULL) {
                printf("File creation failed.\n");
                return 1;
            }
        }
        else if(fileChoice == 3){
            char fileName[100]; // Максимальна довжина назви файлу
            printf("Enter the name of the file to save: ");
            scanf("%s", fileName);
            writeBooksToFile(fileName, books, num_books);
        }
        else{
            printf("Wrong option\n");
        }
    }


        else if (operationChoice == 's' || operationChoice == 'S'){
            printf("1 - by author\n\
            2 - by book title\n\
            3 - by year\n\
            4 - by pages\n\
            5 - by price\n");
            int sortChoice;
            printf("Enter your choice: ");
            scanf(" %d", &sortChoice);
            if(sortChoice == 1)
                sortByAuthor(&head);

            else if(sortChoice == 2)
                sortByTitle(&head);

            else if(sortChoice == 3)
                sortByYear(&head);

            else if(sortChoice == 4)
                sortByPages(&head);

            else if(sortChoice == 5)
                sortByCost(&head);

            else
                printf("Wrong option\n");

            printf("\nBooks in the linked list after operation:\n");
            printList(head);
        }
        else if (operationChoice == 'a' || operationChoice == 'A') {
            struct book newBook;

            printf("Enter author: ");
            scanf("%s", newBook.author);

            printf("Enter title: ");
            scanf("%s", newBook.title);

            printf("Enter year: ");
            scanf("%d", &newBook.year);

            printf("Enter pages: ");
            scanf("%d", &newBook.pages);

            printf("Enter cost: ");
            scanf("%f", &newBook.cost);

            appendNode(&head, newBook);

            printf("\nBook added successfully!\n");

            printf("\nBooks in the linked list after operation:\n");
            printList(head);

        } else if (operationChoice == 'd' || operationChoice == 'D') {
            int index;
            printf("Enter the index of the book to delete: ");
            scanf("%d", &index);

            deleteBook(&head, index - 1);
            printf("\nBooks in the linked list after operation:\n");
            printList(head);

        } else if (operationChoice != 'q' && operationChoice != 'Q') {
            printf("\nInvalid choice. Please enter 'a', 'd', or 'd'.\n");
        }

    } while (operationChoice != 'q' && operationChoice != 'Q');

    freeList(head);

    if(opened) fclose(file);
    
    return 0;
}
